/*
  Explode.C from Jeffrey C. Moxon, CIS: 73561,1431

  GIF file reader converted to file compression routine.  Original GIF file
  reader is from "Bitmapped Graphics, 2nd Edition" by Steve Rimmer from
  McGraw Hill books.  The original GIF reader source can be purchased from:

  Alchemy Mindworks Inc.
  P.O. Box 500
  Beeton, Ontario
  L0G 1A0, Canada

  Although I wouldn't recomend the book because the code is complex and
  sloppy (like mine is neat!), I spent some time playing with the GIF
  file reader and came up with this.  I don't hold any responsibilities
  for the code, nor do I make any claims to its effectiveness.  It works
  moderately well, is written entirely in C and is totally free - I typed
  in the code from the book, and corrected the 50 some-odd warning messages.

  I would like to see a copy with performance improvements if you manage
  any, and/or anything that you produce using this.  As it is, I put minimal
  effort into this, and it still worked out to be fairly valuable.  I also
  have no understanding of the mechanics behind the routine.

  Have fun,
  -JCM-

  PS,
	The headers and main routine aren't necessary. If you open two files,
	one of which was compressed with the compression routine as *sf and
	run them through the explode routine, the data from the source
	file will be de-compressed into the target file.
*/

#include <conio.h>
#include <process.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <io.h>

int bits2;
int codesize;
int codesize2;
int nextcode;
int thiscode;
int oldtoken;
int current_code;
int old_code;
int bitsleft;
int blocksize;
int bits;

unsigned char *p;
unsigned char *q;
unsigned char b[255];
unsigned char *u;

FILE *sf, *tf;

struct file
	   {
	    struct ftime i_date;        //Original date of file
	    char        name[13];       //Original file name
	    long        initial_size;   //Size of original file
	    long        stored_size;    //Size of compressed file
	    char	reserved[32];	//Future Use
	   } files;


struct head
	   {
	    char        sig[5];
	    int         bytes;
	    int         num_files;
	    char	reserved[32];	//Reserved
	   } header;

int uncompress(void);
void display_percent(int c_per);

void main(int argc, char *argv[])
{
 static struct ftime f_date;
 static int file_number;
 static long pos;

 if (argc < 2)
 {
  printf("Requires name of compressed file.\n");
  return;
 }
 if ((sf = fopen(argv[1], "rb")) == NULL)
 {
  printf("Error opening source file.\n");
  return;
 }

 fread(&header, sizeof(header), 1, sf);

 pos = header.bytes;

 for(file_number = 1;file_number < header.num_files;file_number++)
 {
  fseek(sf, sizeof(header) + sizeof(files)*(file_number-1), SEEK_SET);
  fread(&files, sizeof(files), 1, sf);
  if ((tf = fopen(files.name, "wb")) == NULL)
   printf("Error opening target file.\n");

  f_date = files.i_date;
  setftime(fileno(tf), &f_date);
  printf("\nUncompressing:  %13s", files.name);
  fseek(sf, pos, SEEK_SET);
  uncompress();
  rewind(tf);
  pos += files.stored_size;
 }

 printf("\n");
return;
}

int uncompress(void)
{
 int here;
 static long ten_percent;
 static long c_pos;
 static int c_per;
 static unsigned char firstcodestack[4096];  /* Stack for first codes */
 static unsigned char lastcodestack[4096];   /* Statck for previous code */
 static int codestack[4096];        /* Stack for links */

 static int wordmasktable[] = { 0x0000,0x0001,0x0003,0x0007,
				0x000f,0x001f,0x003f,0x007f,
				0x00ff,0x01ff,0x03ff,0x07ff,
				0x0fff,0x1fff,0x3fff,0x7fff
				};

 here = 0;
 p=q=b;
 bitsleft = 8;

 /* read initial code size */
 bits = fgetc(sf);

 bits2 = 1 << bits;
 nextcode = bits2 + 2;
 codesize2 = 1 << (codesize = bits + 1);
 old_code=oldtoken=-1;

 c_per = 0;  //Current percentage;
 ten_percent = (long) files.initial_size * 0.10;
 c_pos = 0;

   /* loop until something breaks */
   for(;;)
   {
     if(bitsleft==8)
     {
      if(++p >= q && (((blocksize = fgetc(sf)) < 1) ||
       (q=(p=b)+fread(b,1,blocksize,sf))< (b+blocksize)))
       return(0);

      bitsleft = 0;
     }
     thiscode = *p;
     if ((current_code=(codesize+bitsleft)) <= 8)
     {
	*p >>= codesize;
	bitsleft = current_code;
     }
     else
     {
       if(++p >= q && (((blocksize = fgetc(sf)) < 1) ||
	       (q=(p=b)+fread(b,1,blocksize,sf)) < (b+blocksize)))
       return(0);

	     thiscode |= *p << (8 - bitsleft);
	     if(current_code <= 16) *p >>= (bitsleft=current_code-8);
	     else
	     {
		     if(++p >= q &&
		       (((blocksize = fgetc(sf)) < 1) ||
		       (q=(p=b) + fread(b,1,blocksize,sf)) < (b+blocksize)))
		     return(0);

		     thiscode |= *p << (16 - bitsleft);
		     *p >>= (bitsleft = current_code - 16);
	     }
     }
     thiscode &= wordmasktable[codesize];
     current_code = thiscode;

     if(thiscode == (bits2+1)) break;   /* found EOI */
     if(thiscode > nextcode)
     {
      printf("Bad LZW code.\n");
      exit(0);
     }

     if(thiscode == bits2)
     {
	     nextcode = bits2 + 2;
	     codesize2 = 1 << (codesize = (bits + 1));
	     oldtoken = old_code = -1;
	     continue;
     }

     u = firstcodestack;

     if(thiscode==nextcode)
     {
	     if(old_code==-1)
	     {
	      printf("Bad First LZW code.\n");
	      exit(0);
	     }

	     *u++ = oldtoken;
	     thiscode = old_code;
     }

     while (thiscode >= bits2)
     {
	     *u++ = lastcodestack[thiscode];
	     thiscode = codestack[thiscode];
     }

      oldtoken = thiscode;

      do
      {
	if (here)
	{
	 fputc(thiscode, tf);
	 c_pos++;
	 if (c_pos >= ten_percent)
	 {
	  display_percent(c_per+=10);
	  c_pos = 0;
	 }
	}
	if (u <= firstcodestack) break;
       thiscode = *--u;
      } while(1);

      here = 1;

     if(nextcode < 4096 && old_code != -1)
     {
	     codestack[nextcode] = old_code;
	     lastcodestack[nextcode] = oldtoken;
	     if (++nextcode >= codesize2 && codesize < 12)
		 codesize2 = 1 << ++codesize;
     }
     old_code = current_code;
   }


return(1);
}

void display_percent(int c_per)
{
 int x, y;

 if (c_per <= 10) //First trip
  x = wherex() + 2;
 else
  x = wherex() - 4;

  y = wherey();

 gotoxy(x, y);
 printf("%i %", c_per);

return;
}

